@@ -86,6 +86,10 @@ class Adversarial_AE(VAE):
         ), "adversarial_loss_scale must be in [0, 1]"
 
         self.adversarial_loss_scale = self.model_config.adversarial_loss_scale
+        self.reconstruction_loss_scale = self.model_config.reconstruction_loss_scale
+        self.deterministic_posterior = (
+            1 if self.model_config.deterministic_posterior else 0
+        )
 
     def set_discriminator(self, discriminator: BaseDiscriminator) -> None:
         r"""This method is called to set the discriminator network
@@ -121,7 +125,7 @@ class Adversarial_AE(VAE):
 
         mu, log_var = encoder_output.embedding, encoder_output.log_covariance
 
-        std = torch.exp(0.5 * log_var)
+        std = (1 - self.deterministic_posterior) * torch.exp(0.5 * log_var)
         z, _ = self._sample_gauss(mu, std)
         recon_x = self.decoder(z)["reconstruction"]
 
@@ -174,15 +178,15 @@ class Adversarial_AE(VAE):
             F.binary_cross_entropy(
                 gen_adversarial_score, true_labels
             )  # generated are true
-        ) + (1 - self.adversarial_loss_scale) * (recon_loss)
+        ) + (recon_loss * self.reconstruction_loss_scale)
 
         gen_adversarial_score_ = self.discriminator(z.detach()).embedding.flatten()
 
-        discriminator_loss = 0.5 * (
+        discriminator_loss = (
             F.binary_cross_entropy(
                 prior_adversarial_score, true_labels
             )  # prior is true
-        ) + 0.5 * (
+        ) + (
             F.binary_cross_entropy(
                 gen_adversarial_score_, fake_labels
             )  # generated are false
