@@ -79,9 +79,18 @@ class FactorVAE(VAE):
             ModelOutput: An instance of ModelOutput containing all the relevant parameters
 
         """
+        x_in = inputs["data"]
+        if x_in.shape[0] <= 1:
+            raise ArithmeticError(
+                "At least 2 samples in a batch are required for the `FactorVAE` model"
+            )
+
+        idx = torch.randperm(x_in.shape[0])
+        idx_1 = idx[int(x_in.shape[0] / 2) :]
+        idx_2 = idx[: int(x_in.shape[0] / 2)]
 
         # first batch
-        x = inputs["data"]
+        x = inputs["data"][idx_1]
 
         encoder_output = self.encoder(x)
 
@@ -92,7 +101,7 @@ class FactorVAE(VAE):
         recon_x = self.decoder(z)["reconstruction"]
 
         # second batch
-        x_bis = inputs["data_bis"]
+        x_bis = inputs["data"][idx_2]
 
         encoder_output = self.encoder(x_bis)
 
@@ -152,10 +161,14 @@ class FactorVAE(VAE):
         permuted_latent_adversarial_score = self.discriminator(z_bis_permuted)
 
         true_labels = (
-            torch.ones(N, requires_grad=False).type(torch.LongTensor).to(z.device)
+            torch.ones(z_bis_permuted.shape[0], requires_grad=False)
+            .type(torch.LongTensor)
+            .to(z.device)
         )
         fake_labels = (
-            torch.zeros(N, requires_grad=False).type(torch.LongTensor).to(z.device)
+            torch.zeros(z.shape[0], requires_grad=False)
+            .type(torch.LongTensor)
+            .to(z.device)
         )
 
         TC_permuted = F.cross_entropy(
@@ -170,6 +183,79 @@ class FactorVAE(VAE):
             (discriminator_loss).mean(dim=0),
         )
 
+    def reconstruct(self, inputs: torch.Tensor):
+        """This function returns the reconstructions of given input data.
+
+        Args:
+            inputs (torch.Tensor): The inputs data to be reconstructed of shape [B x input_dim]
+            ending_inputs (torch.Tensor): The starting inputs in the interpolation of shape
+
+        Returns:
+            torch.Tensor: A tensor of shape [B x input_dim] containing the reconstructed samples.
+        """
+        encoder_output = self.encoder(inputs)
+
+        mu, log_var = encoder_output.embedding, encoder_output.log_covariance
+
+        std = torch.exp(0.5 * log_var)
+        z, _ = self._sample_gauss(mu, std)
+        recon_x = self.decoder(z)["reconstruction"]
+        return recon_x
+
+    def interpolate(
+        self,
+        starting_inputs: torch.Tensor,
+        ending_inputs: torch.Tensor,
+        granularity: int = 10,
+    ):
+        """This function performs a linear interpolation in the latent space of the autoencoder
+        from starting inputs to ending inputs. It returns the interpolation trajectories.
+
+        Args:
+            starting_inputs (torch.Tensor): The starting inputs in the interpolation of shape
+                [B x input_dim]
+            ending_inputs (torch.Tensor): The starting inputs in the interpolation of shape
+                [B x input_dim]
+            granularity (int): The granularity of the interpolation.
+
+        Returns:
+            torch.Tensor: A tensor of shape [B x granularity x input_dim] containing the
+            interpolation trajectories.
+        """
+        assert starting_inputs.shape[0] == ending_inputs.shape[0], (
+            "The number of starting_inputs should equal the number of ending_inputs. Got "
+            f"{starting_inputs.shape[0]} sampler for starting_inputs and {ending_inputs.shape[0]} "
+            "for endinging_inputs."
+        )
+
+        encoder_output = self.encoder(starting_inputs)
+        mu, log_var = encoder_output.embedding, encoder_output.log_covariance
+        std = torch.exp(0.5 * log_var)
+        starting_z, _ = self._sample_gauss(mu, std)
+
+        encoder_output = self.encoder(ending_inputs)
+        mu, log_var = encoder_output.embedding, encoder_output.log_covariance
+        std = torch.exp(0.5 * log_var)
+        ending_z, _ = self._sample_gauss(mu, std)
+
+        t = torch.linspace(0, 1, granularity).to(starting_inputs.device)
+        intep_line = (
+            torch.kron(
+                starting_z.reshape(starting_z.shape[0], -1), (1 - t).unsqueeze(-1)
+            )
+            + torch.kron(ending_z.reshape(ending_z.shape[0], -1), t.unsqueeze(-1))
+        ).reshape((starting_z.shape[0] * t.shape[0],) + (starting_z.shape[1:]))
+
+        decoded_line = self.decoder(intep_line).reconstruction.reshape(
+            (
+                starting_inputs.shape[0],
+                t.shape[0],
+            )
+            + (starting_inputs.shape[1:])
+        )
+
+        return decoded_line
+
     def _sample_gauss(self, mu, std):
         # Reparametrization trick
         # Sample N(0, I)
