@@ -5,9 +5,12 @@ from copy import deepcopy
 from typing import Any, Dict, List, Optional
 
 import torch
+import torch.distributed as dist
 import torch.optim as optim
-from torch.optim.lr_scheduler import ReduceLROnPlateau
+import torch.optim.lr_scheduler as lr_scheduler
+from torch.nn.parallel import DistributedDataParallel as DDP
 from torch.utils.data import DataLoader
+from torch.utils.data.distributed import DistributedSampler
 
 from ...customexception import ModelError
 from ...data.datasets import BaseDataset
@@ -45,12 +48,6 @@ class BaseTrainer:
             parameters used for training. If None, a basic training instance of
             :class:`BaseTrainerConfig` is used. Default: None.
 
-        optimizer (~torch.optim.Optimizer): An instance of `torch.optim.Optimizer` used for
-            training. If None, a :class:`~torch.optim.Adam` optimizer is used. Default: None.
-
-        scheduler (~torch.optim.lr_scheduler): An instance of `torch.optim.Optimizer` used for
-            training. If None, a :class:`~torch.optim.Adam` optimizer is used. Default: None.
-
         callbacks (List[~pythae.trainers.training_callbacks.TrainingCallbacks]):
             A list of callbacks to use during training.
     """
@@ -61,8 +58,6 @@ class BaseTrainer:
         train_dataset: BaseDataset,
         eval_dataset: Optional[BaseDataset] = None,
         training_config: Optional[BaseTrainerConfig] = None,
-        optimizer: Optional[torch.optim.Optimizer] = None,
-        scheduler: Optional[torch.optim.lr_scheduler.LambdaLR] = None,
         callbacks: List[TrainingCallback] = None,
     ):
 
@@ -73,46 +68,43 @@ class BaseTrainer:
             output_dir = "dummy_output_dir"
             training_config.output_dir = output_dir
 
-        if not os.path.exists(training_config.output_dir):
-            os.makedirs(training_config.output_dir)
-            logger.info(
-                f"Created {training_config.output_dir} folder since did not exist.\n"
+        self.training_config = training_config
+        self.model_config = model.model_config
+        self.model_name = model.model_name
+
+        # for distributed training
+        self.world_size = self.training_config.world_size
+        self.local_rank = self.training_config.local_rank
+        self.rank = self.training_config.rank
+        self.dist_backend = self.training_config.dist_backend
+
+        if self.world_size > 1:
+            self.distributed = True
+        else:
+            self.distributed = False
+
+        if self.distributed:
+            device = self._setup_devices()
+
+        else:
+            device = (
+                "cuda"
+                if torch.cuda.is_available() and not self.training_config.no_cuda
+                else "cpu"
             )
 
-        self.training_config = training_config
-
-        set_seed(self.training_config.seed)
-
-        device = (
-            "cuda"
-            if torch.cuda.is_available() and not training_config.no_cuda
-            else "cpu"
-        )
+        self.device = device
 
         # place model on device
         model = model.to(device)
         model.device = device
 
-        # set optimizer
-        if optimizer is None:
-            optimizer = self.set_default_optimizer(model)
-
-        else:
-            optimizer = self._set_optimizer_on_device(optimizer, device)
-
-        # set scheduler
-        if scheduler is None:
-            scheduler = self.set_default_scheduler(model, optimizer)
+        if self.distributed:
+            model = DDP(model, device_ids=[self.local_rank])
 
         self.train_dataset = train_dataset
         self.eval_dataset = eval_dataset
 
-        self.model = model
-        self.optimizer = optimizer
-        self.scheduler = scheduler
-
-        self.device = device
-
         # Define the loaders
         train_loader = self.get_train_dataloader(train_dataset)
 
@@ -128,53 +120,176 @@ class BaseTrainer:
 
         self.train_loader = train_loader
         self.eval_loader = eval_loader
+        self.callbacks = callbacks
 
-        if callbacks is None:
-            callbacks = [TrainingCallback()]
+        # run sanity check on the model
+        self._run_model_sanity_check(model, train_loader)
 
-        self.callback_handler = CallbackHandler(
-            callbacks=callbacks, model=model, optimizer=optimizer, scheduler=scheduler
-        )
+        if self.is_main_process:
+            logger.info("Model passed sanity check !\n" "Ready for training.\n")
 
-        self.callback_handler.add_callback(ProgressBarCallback())
-        self.callback_handler.add_callback(MetricConsolePrinterCallback())
+        self.model = model
+
+    @property
+    def is_main_process(self):
+        if self.rank == 0 or self.rank == -1:
+            return True
+        else:
+            return False
+
+    def _setup_devices(self):
+        """Sets up the devices to perform distributed training."""
+
+        if dist.is_available() and dist.is_initialized() and self.local_rank == -1:
+            logger.warning(
+                "torch.distributed process group is initialized, but local_rank == -1. "
+            )
+        if self.training_config.no_cuda:
+            self._n_gpus = 0
+            device = "cpu"
+
+        else:
+            torch.cuda.set_device(self.local_rank)
+            device = torch.device("cuda", self.local_rank)
+
+            if not dist.is_initialized():
+                dist.init_process_group(
+                    backend=self.dist_backend,
+                    init_method="env://",
+                    world_size=self.world_size,
+                    rank=self.rank,
+                )
+
+        return device
 
     def get_train_dataloader(
         self, train_dataset: BaseDataset
     ) -> torch.utils.data.DataLoader:
-
+        if self.distributed:
+            train_sampler = DistributedSampler(
+                train_dataset, num_replicas=self.world_size, rank=self.rank
+            )
+        else:
+            train_sampler = None
         return DataLoader(
             dataset=train_dataset,
-            batch_size=self.training_config.batch_size,
-            shuffle=True,
+            batch_size=self.training_config.per_device_train_batch_size,
+            num_workers=self.training_config.train_dataloader_num_workers,
+            shuffle=(train_sampler is None),
+            sampler=train_sampler,
         )
 
     def get_eval_dataloader(
         self, eval_dataset: BaseDataset
     ) -> torch.utils.data.DataLoader:
+        if self.distributed:
+            eval_sampler = DistributedSampler(
+                eval_dataset, num_replicas=self.world_size, rank=self.rank
+            )
+        else:
+            eval_sampler = None
         return DataLoader(
             dataset=eval_dataset,
-            batch_size=self.training_config.batch_size,
-            shuffle=False,
+            batch_size=self.training_config.per_device_eval_batch_size,
+            num_workers=self.training_config.eval_dataloader_num_workers,
+            shuffle=(eval_sampler is None),
+            sampler=eval_sampler,
         )
 
-    def set_default_optimizer(self, model: BaseAE) -> torch.optim.Optimizer:
+    def set_optimizer(self):
+        optimizer_cls = getattr(optim, self.training_config.optimizer_cls)
 
-        optimizer = optim.Adam(
-            model.parameters(), lr=self.training_config.learning_rate
+        if self.training_config.optimizer_params is not None:
+            optimizer = optimizer_cls(
+                self.model.parameters(),
+                lr=self.training_config.learning_rate,
+                **self.training_config.optimizer_params,
+            )
+        else:
+            optimizer = optimizer_cls(
+                self.model.parameters(), lr=self.training_config.learning_rate
+            )
+
+        self.optimizer = optimizer
+
+    def set_scheduler(self):
+        if self.training_config.scheduler_cls is not None:
+            scheduler_cls = getattr(lr_scheduler, self.training_config.scheduler_cls)
+
+            if self.training_config.scheduler_params is not None:
+                scheduler = scheduler_cls(
+                    self.optimizer, **self.training_config.scheduler_params
+                )
+            else:
+                scheduler = scheduler_cls(self.optimizer)
+
+        else:
+            scheduler = None
+
+        self.scheduler = scheduler
+
+    def _set_output_dir(self):
+        # Create folder
+        if not os.path.exists(self.training_config.output_dir) and self.is_main_process:
+            os.makedirs(self.training_config.output_dir, exist_ok=True)
+            logger.info(
+                f"Created {self.training_config.output_dir} folder since did not exist.\n"
+            )
+
+        self._training_signature = (
+            str(datetime.datetime.now())[0:19].replace(" ", "_").replace(":", "-")
         )
 
-        return optimizer
-
-    def set_default_scheduler(
-        self, model: BaseAE, optimizer: torch.optim.Optimizer
-    ) -> torch.optim.lr_scheduler:
-
-        scheduler = optim.lr_scheduler.ReduceLROnPlateau(
-            optimizer, factor=0.5, patience=10, verbose=True
+        training_dir = os.path.join(
+            self.training_config.output_dir,
+            f"{self.model_name}_training_{self._training_signature}",
         )
 
-        return scheduler
+        self.training_dir = training_dir
+
+        if not os.path.exists(training_dir) and self.is_main_process:
+            os.makedirs(training_dir, exist_ok=True)
+            logger.info(
+                f"Created {training_dir}. \n"
+                "Training config, checkpoints and final model will be saved here.\n"
+            )
+
+    def _get_file_logger(self, log_output_dir):
+        log_dir = log_output_dir
+
+        # if dir does not exist create it
+        if not os.path.exists(log_dir) and self.is_main_process:
+            os.makedirs(log_dir, exist_ok=True)
+            logger.info(f"Created {log_dir} folder since did not exists.")
+            logger.info("Training logs will be recodered here.\n")
+            logger.info(" -> Training can be monitored here.\n")
+
+        # create and set logger
+        log_name = f"training_logs_{self._training_signature}"
+
+        file_logger = logging.getLogger(log_name)
+        file_logger.setLevel(logging.INFO)
+        f_handler = logging.FileHandler(
+            os.path.join(log_dir, f"training_logs_{self._training_signature}.log")
+        )
+        f_handler.setLevel(logging.INFO)
+        file_logger.addHandler(f_handler)
+
+        # Do not output logs in the console
+        file_logger.propagate = False
+
+        return file_logger
+
+    def _setup_callbacks(self):
+        if self.callbacks is None:
+            self.callbacks = [TrainingCallback()]
+
+        self.callback_handler = CallbackHandler(
+            callbacks=self.callbacks, model=self.model
+        )
+
+        self.callback_handler.add_callback(ProgressBarCallback())
+        self.callback_handler.add_callback(MetricConsolePrinterCallback())
 
     def _run_model_sanity_check(self, model, loader):
         try:
@@ -221,12 +336,13 @@ class BaseTrainer:
                     cuda_inputs[key] = inputs[key].cuda()
 
                 else:
-                    cuda_inputs = inputs[key]
+                    cuda_inputs[key] = inputs[key]
             inputs_on_device = cuda_inputs
 
         return inputs_on_device
 
     def _optimizers_step(self, model_output=None):
+
         loss = model_output.loss
 
         self.optimizer.zero_grad()
@@ -234,12 +350,32 @@ class BaseTrainer:
         self.optimizer.step()
 
     def _schedulers_step(self, metrics=None):
-        if isinstance(self.scheduler, ReduceLROnPlateau):
+        if self.scheduler is None:
+            pass
+
+        elif isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau):
             self.scheduler.step(metrics)
 
         else:
             self.scheduler.step()
 
+    def prepare_training(self):
+        """Sets up the trainer for training"""
+        # set random seed
+        set_seed(self.training_config.seed)
+
+        # set optimizer
+        self.set_optimizer()
+
+        # set scheduler
+        self.set_scheduler()
+
+        # create folder for saving
+        self._set_output_dir()
+
+        # set callbacks
+        self._setup_callbacks()
+
     def train(self, log_output_dir: str = None):
         """This function is the main training function
 
@@ -247,72 +383,37 @@ class BaseTrainer:
             log_output_dir (str): The path in which the log will be stored
         """
 
+        self.prepare_training()
+
         self.callback_handler.on_train_begin(
-            training_config=self.training_config, model_config=self.model.model_config
+            training_config=self.training_config, model_config=self.model_config
         )
 
-        # run sanity check on the model
-        self._run_model_sanity_check(self.model, self.train_loader)
-
-        logger.info("Model passed sanity check !\n")
-
-        self._training_signature = (
-            str(datetime.datetime.now())[0:19].replace(" ", "_").replace(":", "-")
-        )
-
-        training_dir = os.path.join(
-            self.training_config.output_dir,
-            f"{self.model.model_name}_training_{self._training_signature}",
-        )
-
-        self.training_dir = training_dir
-
-        if not os.path.exists(training_dir):
-            os.makedirs(training_dir)
-            logger.info(
-                f"Created {training_dir}. \n"
-                "Training config, checkpoints and final model will be saved here.\n"
-            )
-
         log_verbose = False
 
+        msg = (
+            f"Training params:\n - max_epochs: {self.training_config.num_epochs}\n"
+            " - per_device_train_batch_size: "
+            f"{self.training_config.per_device_train_batch_size}\n"
+            " - per_device_eval_batch_size: "
+            f"{self.training_config.per_device_eval_batch_size}\n"
+            f" - checkpoint saving every: {self.training_config.steps_saving}\n"
+            f"Optimizer: {self.optimizer}\n"
+            f"Scheduler: {self.scheduler}\n"
+        )
+
+        if self.is_main_process:
+            logger.info(msg)
+
         # set up log file
-        if log_output_dir is not None:
-            log_dir = log_output_dir
+        if log_output_dir is not None and self.is_main_process:
             log_verbose = True
+            file_logger = self._get_file_logger(log_output_dir=log_output_dir)
 
-            # if dir does not exist create it
-            if not os.path.exists(log_dir):
-                os.makedirs(log_dir)
-                logger.info(f"Created {log_dir} folder since did not exists.")
-                logger.info("Training logs will be recodered here.\n")
-                logger.info(" -> Training can be monitored here.\n")
+            file_logger.info(msg)
 
-            # create and set logger
-            log_name = f"training_logs_{self._training_signature}"
-
-            file_logger = logging.getLogger(log_name)
-            file_logger.setLevel(logging.INFO)
-            f_handler = logging.FileHandler(
-                os.path.join(log_dir, f"training_logs_{self._training_signature}.log")
-            )
-            f_handler.setLevel(logging.INFO)
-            file_logger.addHandler(f_handler)
-
-            # Do not output logs in the console
-            file_logger.propagate = False
-
-            file_logger.info("Training started !\n")
-            file_logger.info(
-                f"Training params:\n - max_epochs: {self.training_config.num_epochs}\n"
-                f" - batch_size: {self.training_config.batch_size}\n"
-                f" - checkpoint saving every {self.training_config.steps_saving}\n"
-            )
-
-            file_logger.info(f"Model Architecture: {self.model}\n")
-            file_logger.info(f"Optimizer: {self.optimizer}\n")
-
-        logger.info("Successfully launched training !\n")
+        if self.is_main_process:
+            logger.info("Successfully launched training !\n")
 
         # set best losses for early stopping
         best_train_loss = 1e10
@@ -360,6 +461,7 @@ class BaseTrainer:
             if (
                 self.training_config.steps_predict is not None
                 and epoch % self.training_config.steps_predict == 0
+                and self.is_main_process
             ):
                 true_data, reconstructions, generations = self.predict(best_model)
 
@@ -378,23 +480,33 @@ class BaseTrainer:
                 self.training_config.steps_saving is not None
                 and epoch % self.training_config.steps_saving == 0
             ):
-                self.save_checkpoint(
-                    model=best_model, dir_path=training_dir, epoch=epoch
-                )
-                logger.info(f"Saved checkpoint at epoch {epoch}\n")
+                if self.is_main_process:
+                    self.save_checkpoint(
+                        model=best_model, dir_path=self.training_dir, epoch=epoch
+                    )
+                    logger.info(f"Saved checkpoint at epoch {epoch}\n")
 
-                if log_verbose:
-                    file_logger.info(f"Saved checkpoint at epoch {epoch}\n")
+                    if log_verbose:
+                        file_logger.info(f"Saved checkpoint at epoch {epoch}\n")
 
             self.callback_handler.on_log(
-                self.training_config, metrics, logger=logger, global_step=epoch
+                self.training_config,
+                metrics,
+                logger=logger,
+                global_step=epoch,
+                rank=self.rank,
             )
 
-        final_dir = os.path.join(training_dir, "final_model")
+        final_dir = os.path.join(self.training_dir, "final_model")
 
-        self.save_model(best_model, dir_path=final_dir)
-        logger.info("Training ended!")
-        logger.info(f"Saved final model in {final_dir}")
+        if self.is_main_process:
+            self.save_model(best_model, dir_path=final_dir)
+
+            logger.info("Training ended!")
+            logger.info(f"Saved final model in {final_dir}")
+
+        if self.distributed:
+            dist.destroy_process_group()
 
         self.callback_handler.on_train_end(self.training_config)
 
@@ -412,6 +524,7 @@ class BaseTrainer:
             training_config=self.training_config,
             eval_loader=self.eval_loader,
             epoch=epoch,
+            rank=self.rank,
         )
 
         self.model.eval()
@@ -426,12 +539,18 @@ class BaseTrainer:
                 with torch.no_grad():
 
                     model_output = self.model(
-                        inputs, epoch=epoch, dataset_size=len(self.eval_loader.dataset)
+                        inputs,
+                        epoch=epoch,
+                        dataset_size=len(self.eval_loader.dataset),
+                        uses_ddp=self.distributed,
                     )
 
             except RuntimeError:
                 model_output = self.model(
-                    inputs, epoch=epoch, dataset_size=len(self.eval_loader.dataset)
+                    inputs,
+                    epoch=epoch,
+                    dataset_size=len(self.eval_loader.dataset),
+                    uses_ddp=self.distributed,
                 )
 
             loss = model_output.loss
@@ -460,6 +579,7 @@ class BaseTrainer:
             training_config=self.training_config,
             train_loader=self.train_loader,
             epoch=epoch,
+            rank=self.rank,
         )
 
         # set model in train model
@@ -472,7 +592,10 @@ class BaseTrainer:
             inputs = self._set_inputs_to_device(inputs)
 
             model_output = self.model(
-                inputs, epoch=epoch, dataset_size=len(self.train_loader.dataset)
+                inputs,
+                epoch=epoch,
+                dataset_size=len(self.train_loader.dataset),
+                uses_ddp=self.distributed,
             )
 
             self._optimizers_step(model_output)
@@ -489,7 +612,10 @@ class BaseTrainer:
             )
 
         # Allows model updates if needed
-        self.model.update()
+        if self.distributed:
+            self.model.module.update()
+        else:
+            self.model.update()
 
         epoch_loss /= len(self.train_loader)
 
@@ -507,7 +633,11 @@ class BaseTrainer:
             os.makedirs(dir_path)
 
         # save model
-        model.save(dir_path)
+        if self.distributed:
+            model.module.save(dir_path)
+
+        else:
+            model.save(dir_path)
 
         # save training config
         self.training_config.save_json(dir_path, "training_config")
@@ -533,13 +663,18 @@ class BaseTrainer:
         )
 
         # save scheduler
-        torch.save(
-            deepcopy(self.scheduler.state_dict()),
-            os.path.join(checkpoint_dir, "scheduler.pt"),
-        )
+        if self.scheduler is not None:
+            torch.save(
+                deepcopy(self.scheduler.state_dict()),
+                os.path.join(checkpoint_dir, "scheduler.pt"),
+            )
 
         # save model
-        model.save(checkpoint_dir)
+        if self.distributed:
+            model.module.save(checkpoint_dir)
+
+        else:
+            model.save(checkpoint_dir)
 
         # save training config
         self.training_config.save_json(checkpoint_dir, "training_config")
@@ -548,17 +683,22 @@ class BaseTrainer:
 
         model.eval()
 
-        # with torch.no_grad():
-
-        inputs = self.eval_loader.dataset[
-            : min(self.eval_loader.dataset.data.shape[0], 10)
-        ]
+        inputs = next(iter(self.eval_loader))
         inputs = self._set_inputs_to_device(inputs)
 
         model_out = model(inputs)
-        reconstructions = model_out.recon_x.cpu().detach()
-        z_enc = model_out.z
+        reconstructions = model_out.recon_x.cpu().detach()[
+            : min(inputs["data"].shape[0], 10)
+        ]
+        z_enc = model_out.z[: min(inputs["data"].shape[0], 10)]
         z = torch.randn_like(z_enc)
-        normal_generation = model.decoder(z).reconstruction.detach().cpu()
+        if self.distributed:
+            normal_generation = model.module.decoder(z).reconstruction.detach().cpu()
+        else:
+            normal_generation = model.decoder(z).reconstruction.detach().cpu()
 
-        return inputs["data"], reconstructions, normal_generation
+        return (
+            inputs["data"][: min(inputs["data"].shape[0], 10)],
+            reconstructions,
+            normal_generation,
+        )
